/* gnome-napster, a GNOME Napster client.
 * Copyright (C) 1999 Evan Martin <eeyem@u.washington.edu>
 */


/* FIXME
 * This entire file is a big messy hack.  It should be thrown away and rewritten.
 */
/* split this into two files(at least) one called download_ui.c and one called
download_thread.c.  download_ui.c should have all of the ui code or GTK
interface; NONE of the actuall code to handel the control or actuall 
recieving of the download. download_thread.c contains all of the control 
and such.  It should export an interface that allows enqueueing, status
updates, starting a download, canceling a download, etc.  */
/* this needs to be done to allow splitting into an eventual libnapster shared
library. */

#include <gnome.h>
#include <glib.h>
#include <pthread.h>
#include <sys/types.h>
#include <errno.h>
#include <unistd.h>
#include <ctype.h> /* for isdigit */

/* gettimeofday */
#include <sys/time.h>

/* open */
#include <sys/stat.h>
#include <fcntl.h>
/* network */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "lnap.h"
#include "download.h"
#include "login.h" /* for the login info (client name) */
#include "host.h" /* to show the host info dialog */
#include "songinfo.h" /* also to show the host info dialog */
#include "gconfig.h" /* download path */
#include "util.h"
/****************************************************
PROTOTYPES
****************************************************/
static void download_cancel_cb(GtkWidget *w, gpointer d);
static void host_info_cb(GtkWidget *w, gpointer d);
/****************************************************
TYPEDEFS
****************************************************/
typedef enum {
	EG_QUEUED, EG_GETINFO, EG_REMOTEQUEUEFULL, EG_READY,
	EG_CONNECTING, EG_REQUESTING, EG_DOWNLOADING,
	EG_FILECOMPLETE, EG_AWAITING_CONNECTION, EG_ERROR
} EGetStatus;
static char *statuslookup[] = {
	"Queued", "Getting info...", "Remotely Queued", "Ready to download",
	"Connecting...", "Requesting file...", "Downloading...",
	"File complete", "Waiting for connection...", "Error downloading?", NULL
};

enum {
	COL_TITLE, COL_HOST, COL_STATUS, COL_COUNT
};
/* titles on the CList */
static char* titles[] = { "File", "Host", "Status" };

/* popup text */
static GnomeUIInfo popup_menu[] = {
/*
	GNOMEUIINFO_ITEM_NONE(N_("_Select All"),
		N_("Select all items in the list"),
		selall_cb),

	GNOMEUIINFO_ITEM_NONE(N_("_Song Info"),
		N_("View information about this download"),
		song_info_cb),
*/
	GNOMEUIINFO_ITEM_NONE(N_("_Cancel"),
		N_("Cancel this download"),
		download_cancel_cb),
	GNOMEUIINFO_ITEM_NONE(N_("_Host Info"),
		N_("View information about this host"),
		host_info_cb),
	GNOMEUIINFO_END
};

typedef struct {
	char host[20];
	int downloadcur;
	int downloadmax;
	GList *ldl;
	GList *lqueue;
	GMutex *mutex;
} shostqueue;

typedef struct {
	/* data returned by napster */
	struct in_addr sin_addr; /* the address of the remote host */
	int port;		/* the port they want us to connect to */
	char title[256];	/* the title of the file */
	char checksum[40];	/* the md5 checksum */
	int conn;		/* the connection type? */

	/* internal data */
	shostqueue *hq;
	pthread_t thread;	/* the thread that is handeling it */
	int row;		/* the row in the clist to display info */
	int shift;		/* how much of the title to save */
	unsigned long filesize; /* size of the file */
	unsigned long filebytes;
	EGetStatus status;	/* the current status of the download */
	char *errormsg;		/* current error message */

	ssong_info *i;

	/* for alternate downloads */
	int sock;		/* the socket to recieve from */
	GCond *wait_cond;	/* wait on a connection */
	GMutex *mutex;

	struct timeval timestamp;

	/* transfer rate calculations */
	long bytestamp;
	double kbps;
} sgetfile;
/*********************************************
GLOBALS
*********************************************/
static GList *hostlist = NULL;
static int downloadcount = 0;
G_LOCK_DEFINE_STATIC(hostlist);
int alt_sockfd = 0;  /* the socket to listen on for alternate downloads*/
#define DOWNLOAD_UPDATE_FREQ 1000
//static int gtag_update = 0;

static GtkWidget *clist;
G_LOCK_DEFINE_STATIC(clist);
/********************************************
PROTOTYPES
********************************************/
static void queuecheckall(void);
static void queuecheck(shostqueue *hq);
static void getinfo(sgetfile *gf);
static shostqueue* hostqueue_find(char *host, int newok);
static sgetfile* getfile_find(char *host, char *title, int newok);
static sgetfile* getfile_find_download(char *host, char *title);
static void downloadcomplete(sgetfile *gf);
static GList *download_remove_from_queue(sgetfile *gf, GList *q, gboolean *worked);

#define TLOCK(x) { FUNCPRINT("lock"); g_mutex_lock(x); }
#define TUNLOCK(x) { FUNCPRINT("unlock"); g_mutex_unlock(x); }
/*******************************************
FUNCTIONS
*******************************************/
/************************************
gboolean download_release(int sock, char *host, char *filename, unsigned long filesize)

Wake up a previously suspended alternate download

sock: the socket for the download
host: the host for the download
filename: the name of the file coming in

RETURNS: true if it found the host
	false otherwise
************************************/
gboolean download_release(int sock, char *host, char *filename, unsigned long filesize)
{
	sgetfile *gf;
	/* find the file description */	
	if((gf = getfile_find_download(host, filename))==NULL)
	{
		FUNCPRINT("Didn't find it");
		return FALSE;
	}
	FUNCPRINT("found host, releasing download");
	/* setup the info */
	TLOCK(gf->mutex);
	gf->sock=sock;
	gf->filesize = filesize;
	/* start it going */
	g_cond_signal(gf->wait_cond);
	TUNLOCK(gf->mutex);
	FUNCPRINT("Download released");
	return TRUE;
}
/************************************
int download_init(void)

Initialize the download area
************************************/
int download_init(void)
{
	/* Setup the alternate download
		listening socket */
	struct sockaddr_in s;
	int port;
	int optval = 1;

	/* get the port out of the config database */
	port = gnome_config_get_int(CONFIG_PREFIX "/Login/DataPort");
	FUNCPRINT("Listening on port %d", port);
	/* make the socket */
	if ((alt_sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
		perror("sock");
		return -1;
	}
	if (setsockopt(alt_sockfd, SOL_SOCKET, SO_REUSEADDR, &optval, sizeof(optval)) < 0) {
                perror("setsockopt");
                return -1;
        }
	if(fcntl(alt_sockfd, F_SETFL, O_NONBLOCK) < 0) 
	{
		perror("fcntl");
		return -1;
	}
	/* bind the socket to the selected port */
	memset(&s, 0, sizeof(struct sockaddr_in));
	s.sin_family = AF_INET;
	s.sin_port = htons(port);
	/* bind the socket to the correct port */
	if(bind(alt_sockfd, 
		(struct sockaddr *)&s, 
		sizeof(struct sockaddr))<0)
	{
		perror("bind");
		return -1;
	}
	/* listen on the socket */
	if(listen(alt_sockfd, 5)<0)
	{
		perror("listen");
		return -1;
	}
	return alt_sockfd;
}
/**********************************
gboolean download_shutdown(void)

shutdown the download are, may eventually
stop all current downloads as well.

RETURNS: TRUE if is successfully shut 
	the download are down
	FALSE otherwise
**********************************/
gboolean download_shutdown(void)
{
	if(alt_sockfd = -1)
	{
		close(alt_sockfd);
		alt_sockfd = -1;
	}
	return TRUE;
}
/***************************************
void downloadproc_exit(void *d)

The exit handeler for the download threads.
Cleans up the sgetfile struct, by calling downloadcomplete().

d:  pointer to the sgetfile to remove
***************************************/
static void downloadproc_exit(void *d)
{
	FUNCPRINT("exiting thread");
	downloadcomplete((sgetfile *)d);
}
/*************************************
static shostqueue *shostqueue_new(char *host)

Add a new host node to the list of queued hosts

host: the name of the host to add

RETURNS: pointer to the new node
*************************************/
static shostqueue* 
shostqueue_new(char *host) {
	shostqueue* hq = g_new0(shostqueue, 1);
	strcpy(hq->host, host);
	hq->mutex = g_mutex_new();
	hq->downloadmax = 2; /* FIXME */
	return hq;
}
/**************************************
static sgetfile *sgetfile_new(char *title)

Makes a new sgetfile structure and sets the title.
All other parts are initialized to 0.

title: the title to initialize the struct with

RETURNS: pointer to the new sgetfile
****************************************/
static sgetfile* 
sgetfile_new(char *title) {
	sgetfile *gf;
	FUNCPRINT("title='%s'", title);
	gf = g_new0(sgetfile, 1);


	strcpy(gf->title, title);
	gf->wait_cond = NULL;
	gf->mutex = NULL;

	return gf;
}

static void 
list_update(int row) {
	sgetfile *gf;
	char buf[1024];
	int hours = 0, minutes = 0, seconds = 0;
	if(row < 0) return;
	gf = (sgetfile*) gtk_clist_get_row_data(GTK_CLIST(clist), row);
	/*FUNCPRINT("row=%d, status=%s", row, statuslookup[gf->status]);*/
	if (gf->status == EG_DOWNLOADING) {
		/* get the total number of seconds left to download */
		if(gf->kbps)
			seconds = ((gf->filesize - gf->filebytes)/1024) / gf->kbps;
		else seconds = 65535;/*FIXME: Magic number, need to fix */
		/* split it up to understandable units */
		hours = seconds / 3600;
		minutes = (seconds % 3600) / 60;
		seconds = seconds % 60;

		/*sprintf(buf, "%.2f/%.2fMb: %.1f%% (%.2fKb/sec) %2d:%2d:%2d remaining", 
		
			gf->filebytes/(float)1000000, gf->filesize/(float)1000000, 
				100.0*gf->filebytes/(float)gf->filesize, gf->kbps,
				hours, minutes, seconds); */
		sprintf(buf, "%.1f%% done with %02d:%02d:%02d left for %.2f/%.2fMb @ %.2fKb/sec",
			100.0*(gf->filebytes/(float)gf->filesize), 
			hours, minutes, seconds,
			gf->filebytes/1000000.0, 
			gf->filesize/1000000.0, gf->kbps);
		gtk_clist_set_text(GTK_CLIST(clist), row, COL_STATUS, buf);
	} else if (gf->status == EG_ERROR) {
		gtk_clist_set_text(GTK_CLIST(clist), row, COL_STATUS, gf->errormsg);
	} else {
		gtk_clist_set_text(GTK_CLIST(clist), row, COL_STATUS, statuslookup[gf->status]);
	}
	gtk_widget_queue_draw(clist);
}

void download_enqueue(/* ssong_info *i*/ char *host, char *title, int shift) {
	sgetfile *gf;
	gchar* add[COL_COUNT] = { "", "", "" };
	int row;

	/* FUNCPRINT("'%s', '%s'", i->host, i->title); */

	printcf("Adding '%s' to download queue...\r\n", title);
	/* create the gf and hostqueue if it doesn't already exist */
	gf = getfile_find(host, title, TRUE); 
	gf->shift = shift;
	

	add[COL_TITLE] = short_path(gf->title, gf->shift);
	add[COL_HOST] = host;

	G_LOCK(clist);
	row = gtk_clist_append(GTK_CLIST(clist), add);
	gtk_clist_set_row_data(GTK_CLIST(clist), row, gf);
	gf->row = row;

	list_update(row);
	G_UNLOCK(clist);

	getinfo(gf);
}

static void 
getinfo(sgetfile *gf) {
	char buf[1024];

	gf->status = EG_GETINFO;
	G_LOCK(clist);
	list_update( gtk_clist_find_row_from_data(GTK_CLIST(clist), gf) );
	G_UNLOCK(clist);
	FUNCPRINT("gf=%p", gf);
	sprintf(buf, "%s \"%s\"", gf->hq->host, gf->title);
	FUNCPRINT("pack: '%s'\n", buf);
	nap_sendpacket(nap_socknap, NM_FILEINFO_REQUEST, strlen(buf)+1, buf);
}

void download_getinfo_cb(char *text) {
	sgetfile *gf;
	char host[20];
	char title[256];

	FUNCPRINT("packet='%s'", text);

	/* first grab only the host and title */
	sscanf(text, "%s %*u %*d \"%[^\"]\"", host, title);
	/* and look it up */
	gf = getfile_find(host, title, FALSE);
/*	g_assert(gf != NULL); */
	if(gf == NULL) 
	{
		FUNCPRINT("Got info for a cancled download");
		/* it has been canceled */
		return;
	}
	FUNCPRINT("found the struct for the download");	
	/* then scan the rest of the data into the found gf */
	sscanf(text, "%s %u %d \"%[^\"]\" %s %d",
			host, &gf->sin_addr.s_addr, &gf->port, gf->title, gf->checksum, &gf->conn);
	gf->sin_addr.s_addr = nap_naptohl(gf->sin_addr.s_addr);

	gf->status = EG_READY;
	FUNCPRINT("Updating clist...");
	G_LOCK(clist);
	T( list_update(gtk_clist_find_row_from_data(GTK_CLIST(clist), gf)); )
	G_UNLOCK(clist);
	FUNCPRINT("Checking the queue...");
	queuecheckall();
}

void download_remotequeuefull_cb(char *text) {
	sgetfile *gf;
	char host[20];
	char title[256];

	FUNCPRINT("packet='%s'", text);
	/* first grab only the host and title */
	sscanf(text, "%s \"%[^\"]\"", host, title);
	/* and look it up */
	gf = getfile_find(host, title, FALSE);
/*	g_assert(gf != NULL);*/
	if(gf == NULL)
	{
		FUNCPRINT("This has been canceled");
		return;
	}
	FUNCPRINT("Found struct for this download");
	sscanf(text, "%s \"%[^\"]\" %*d %*d",
			host, gf->title);
	FUNCPRINT("Updating clist...");
	gf->status = EG_REMOTEQUEUEFULL;
	G_LOCK(clist);
	T( list_update( gtk_clist_find_row_from_data(GTK_CLIST(clist), gf)); )
	G_UNLOCK(clist);
	FUNCPRINT("done");
}
/********************************************
static void downloadcomplete(sgetfile *gf)

remove the gf from the download queue(ldl) because
it is done downloading

gf: the pointer to the sgetfile struct to remove
********************************************/
static void downloadcomplete(sgetfile *gf) {
	shostqueue *hq = gf->hq;
	GList *li = NULL;
	GList *newhead = NULL;
	gboolean worked = FALSE;
	/* notify host queue that download has completed */
	FUNCPRINT("decrementing download count...");

	G_LOCK(hostlist);
	downloadcount--;
	G_UNLOCK(hostlist);

	TLOCK(hq->mutex);

	/* remove the node */	
	/*newhead = download_remove_from_queue(gf, hq->ldl, &worked);*/

	/* update the head */
	/*if(worked) hq->ldl = newhead;
	else return;*/

	hq->downloadcur--;

	TUNLOCK(hq->mutex);
	FUNCPRINT("queuecheck...");
	queuecheck(gf->hq);
}
/***********************************************
static gboolean download_remove_from_queue(sgetfile *gf, GList **q)

remove the specified node from the specified list 

gf: the data to remove
q: the list to search for it in

RETURNS: 0xdead if it wasn't able to find the node
	otherwise the pointer to the new head of the q list
***********************************************/
static GList *download_remove_from_queue(sgetfile *gf, GList *q, gboolean *worked)
{
	GList *li = NULL;
	GList *queue = q;
	sgetfile *found;
	/* remove from download list */
	if((li = g_list_find(queue, gf)) == NULL)
	{
		FUNCPRINT("The node wasn't here");
		/* it doesn't exist here */
		*worked = FALSE;
		return NULL;
	}
	FUNCPRINT("Found the node, deleteing...");
	queue = g_list_remove_link(queue, li);

	found = li->data;

	/* and clean up allocated members */
	if(found->wait_cond != NULL)
		g_cond_free(found->wait_cond);

	if(found->mutex != NULL)
		g_mutex_free(found->mutex);


	g_list_free_1(li);
	*worked = TRUE;
	return queue;
}
/***********************************************
int download_connect(struct sockaddr_in *sock_info)

make the connection using the info  in  sock_info

RETURNS: socket descriptor
***********************************************/
static int download_connect(struct sockaddr_in *s)
{
	int sockfd;
	int port;
	/* and make the connection */
        FUNCPRINT("connecting %s:%d...\n", inet_ntoa(s->sin_addr), port=ntohs(s->sin_port));
	if(port)
	{
		/* make to socket for the connection to the other client */
		if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
			perror("sock");
			pthread_testcancel(); /* check for a cancel */
			return 0;
		}
		pthread_testcancel(); /* check for a cancel */

		/* they _supposedly_ aren't behind a firewall */
	        if (connect(sockfd, (struct sockaddr*)s, sizeof(struct sockaddr_in)) < 0) {
	                perror("connect");
			pthread_testcancel(); /* check for a cancel */
			/* connection was refused need to try the other way */
			return -1;
       		}
		pthread_testcancel(); /* check for a cancel */
		return sockfd;
	}
	else
	{
		FUNCPRINT("Remote host behind a firewall");
		/* they are behind a firewall */
		return -1;
	}
}
/**********************************************
gboolean download_get_data(int sock, sgetfile *gf)

recieve the file and place it into the file on disk

sock: the socket to recieve from
gf:   the get file struct with the info

RETURNS: true if it completed successfully 
	false otherwise
**********************************************/
gboolean download_get_data(int sock, sgetfile *gf)
{
	const int buflen=2048;
	guchar buf[buflen], dlpath[buflen];
	int i;
	char *configpath=NULL;
	int filefd;
	int len;
	struct timeval ts;
	FUNCPRINT("Recieving file: %s", gf->title);
	configpath = gnome_config_get_string(CONFIG_PREFIX "/Paths/Download");
	sprintf(dlpath, "%s/pending/%s",
			configpath ? configpath : "", /* avoid NULL */
			short_path(gf->title, gf->shift));

	for (i = 0; dlpath[i] != 0; i++) {
		if (dlpath[i] == '\\' || dlpath[i] == '/') {
			/* handle subdirs */
			dlpath[i] = 0;
			mkdir(dlpath, 0777);
			/*FUNCPRINT("making dir '%s'\n", dlpath);*/
			dlpath[i] = '/';
		}
	}
	if (configpath) g_free(configpath);
	FUNCPRINT("opening %s", dlpath);

	if ((filefd = open(dlpath, O_WRONLY|O_CREAT, 0666)) < 0) {
		perror("open");
		return FALSE;
	}
	pthread_testcancel(); /* check for a cancel */
	
	gf->status = EG_DOWNLOADING;

	T( list_update( gtk_clist_find_row_from_data(GTK_CLIST(clist), gf) ); )

	gettimeofday(&gf->timestamp, NULL);
	len = recv(sock, buf, buflen, 0); /* get the first part */
	pthread_testcancel(); /* check for a cancel */
	do { /* write, then read, because we have the first bytes from above */
		write(filefd, buf, len);
		gf->filebytes += len;

		/* and now, some kb/sec calculation.
		* isn't it sad that the calculation code is 8 times the lines of the actual dl?
		* FIXME current code stops updating when no data comes in;
		* it should show it's getting 0kb/sec somehow. */
		//if (gf->filebytes > 10000) break;  //FIXME download hack */
		gettimeofday(&ts, NULL);
		if ((ts.tv_sec - gf->timestamp.tv_sec) > 1) {
			double deltat, deltatu, deltab;
			deltab = (double)gf->filebytes - gf->bytestamp;
			deltat = (double)ts.tv_sec - gf->timestamp.tv_sec;
			deltatu = (double)ts.tv_usec - gf->timestamp.tv_usec;
			gf->kbps = deltab / ((deltat * 1000.0) + (deltatu / 1000.0));
			T( list_update( gtk_clist_find_row_from_data(GTK_CLIST(clist), gf)); )
			gf->bytestamp = gf->filebytes;
			gf->timestamp = ts;
		}
		if(gf->filebytes >= gf->filesize) break;  /* we got the whole thing */
		pthread_testcancel(); /* check for a cancel */
	} while ((len = recv(sock, buf, buflen, 0)) > 0);

	/* note that you get a "connection reset by peer" at the end
	* of a _successful_transfer_! so it's only an error if the file
	* is incomplete, _and_ there was a recv() error
	*/
	if ((gf->filebytes < gf->filesize) && (len == -1)) {
		FUNCPRINT("Error Downloading");
		perror("recv");
		close(filefd);
		return FALSE;
	}

	/* we made it! */
	close(filefd);
	return TRUE;

}
/***********************************************
gboolean standard_download(int sock, sgetfile *gf)

standard connect out to remote host download

sock: the socket to use 
gf:   pointer to the file info to get
***********************************************/
static gboolean standard_download(int sock, sgetfile *gf)
{
	const int buflen=2048;
	guchar buf[buflen], dlpath[buflen];
	int i;
	char *configpath=NULL;
	int filefd;
	int len;
	struct timeval ts;

	 /* get "mysterious number" */
        FUNCPRINT("reading mysterious number 1");
        if (recv(sock, buf, 1, 0) < 0) /* retrieve the initial '1' */
                perror("recv");
        buf[1] = 0;
        FUNCPRINT("number: %s", buf);

        /* the GET command must be two packets... ? */
        sprintf(buf, "GET");
        if (send(sock, buf, strlen(buf), 0) < 0) {
                perror("send");
		return FALSE;
        }
        sprintf(buf, "%s \"%s\" 0", g_logininfo.user, gf->title); /* FIXME: last param (0) */
        FUNCPRINT("sending request...[%s]", buf);
        if (send(sock, buf, strlen(buf), 0) < 0) {
                perror("send");
		return FALSE;
        }
	/* start recieving */
        for (i = 0; i < buflen; i++) {
		/* Peek it */
                recv(sock, buf+i, 1, MSG_PEEK);
                if (!isdigit(buf[i])) break; /* first non-digit char is the first byte of the mp3 */
		/* Read it */
		recv(sock, buf+i, 1, 0);
        }
	/* Terminate the string */
        buf[i] = '\0';
        FUNCPRINT("they say '%s'", buf);
	/* check for an error */
        if (strncmp(buf, "FILE NOT SHARED", 15) == 0) {
                FUNCPRINT("file not shared?!");
                gf->status = EG_ERROR;
                gf->errormsg = "File not shared!?";
                downloadcomplete(gf);
                return FALSE;
        }
	if(!strncmp(buf, "INVALID FILE", 12))
	{
		gf->status = EG_ERROR;
		gf->errormsg = "Invalid file name!?";
		FUNCPRINT("%s", gf->errormsg);
		downloadcomplete(gf);
		return FALSE;
	}
	/* convert the string to a long and record the size */
        gf->filesize = atol(buf);
        FUNCPRINT("file is %ld bytes", gf->filesize);

	/* get the file and return */
	return download_get_data(sock, gf);
}
/***********************************************
void download_accept_alternate()

wait for a connection then return the socket for it
***********************************************/
void download_accept_alternate(sgetfile *gf)
{
	GTimeVal pause = {0, 1};
	FUNCPRINT("Awaiting Connection");
	gf->status = EG_AWAITING_CONNECTION;
	T( list_update( gtk_clist_find_row_from_data(GTK_CLIST(clist), gf)); )
	
	gf->wait_cond = g_cond_new();
/* then wait for the recieving process to unlock us */
	gf->sock=0;
	while(!gf->sock)
	{
		pthread_testcancel(); /* check for a cancel */
		if(g_cond_timed_wait(gf->wait_cond, gf->mutex, &pause) && gf->sock) break;
	}
}
/***********************************************
gboolean alternate_download(sgetfile *gf)

alternate download, wait for the remote host to connect to us

sock: the socket to use
gf:   pointer to the file info to get
***********************************************/
static gboolean alternate_download(sgetfile *gf)
{
	#define BUF_LEN 128
	char buf[BUF_LEN];
	int sock;
	struct sockaddr_in s;
	int s_len;
	/* make the connection and lock it */
	gf->mutex = g_mutex_new();
	/* get the socket */
	download_accept_alternate(gf);
	FUNCPRINT("They connected!");
	sock = gf->sock;
	
	/* send out the offset to start at
		0 for now */
	buf[0]='0';
	buf[1]='\0';
	if(send(sock, buf, strlen(buf), 0) < 0)
	{
		perror("send");
		gf->sock=-1;
		pthread_testcancel(); /* check for a cancel */
		return FALSE;
	} 
	/* recv the data */
	if(!download_get_data(sock, gf)) 
	{
		return FALSE;	
	}
	FUNCPRINT("Downloaded OK");
	return TRUE;
}
/************************************************
static void *downlaodproc(void *d)

download thread

d: the pointer to the struct to save the file info in
************************************************/
static void* 
downloadproc(void* d) {
	const int buflen = 2048; /* just needs to be greater than MTU; 1500, i guess */
	int sockfd = -1, i, ok;
	char *configpath;
	char dlpath[buflen], finalpath[buflen];
	struct sockaddr_in s;
	sgetfile *gf = (sgetfile*)d;

/* setup the pthread exit handler */
pthread_cleanup_push(downloadproc_exit, d);


	/*FUNCPRINT("gf=%p", gf);*/
for (ok = FALSE; ok != TRUE; ok = TRUE) { /* a simplistic try..catch block */
	/* set up the addr_in(?) data struct */
	memset(&s, 0, sizeof(struct sockaddr_in));
	s.sin_family = AF_INET;
	s.sin_port = htons(gf->port);
	s.sin_addr.s_addr = gf->sin_addr.s_addr;

	/* and make the connection */
	gf->status = EG_CONNECTING;
	T( list_update( gtk_clist_find_row_from_data(GTK_CLIST(clist), gf)); )
	if((sockfd=download_connect(&s))<=0)
	{
		pthread_testcancel();
		if(sockfd!=0)
		{
			/* they are behind a firewall */
			FUNCPRINT("Behind a firewall; trying alternate download");
			if(behind_firewall())
			{
				FUNCPRINT("So are we! Can't do it!");
				break;
			}
			/* send the message to the server to have the remote
				host push out a connection */
			sprintf(dlpath, "%s \"%s\"", gf->hq->host, gf->title);
			nap_sendpacket(nap_socknap, NM_ALT_DOWNLOAD_REQUEST, 
					strlen(dlpath), dlpath);
			FUNCPRINT("Sent info, now waiting");
			if(!alternate_download(gf)) 
			{
				FUNCPRINT("Error on alt download");
				break;
			}
			else 
			{
				FUNCPRINT("A OK");
				sockfd=gf->sock;
			}
		}
		else
		{
			FUNCPRINT("Unable to make socket");
			break;
		}
	}
	/* this is just a standard download */
	else 
	{
		pthread_testcancel(); 
		if(!standard_download(sockfd, gf)) break; 
	}

}/* end of for(ok) loop */

	if (ok) {
		/* The download completed ok */
		gf->status = EG_FILECOMPLETE;
		FUNCPRINT("file complete.");
		configpath = gnome_config_get_string(CONFIG_PREFIX "/Paths/Download");
		sprintf(finalpath, "%s/%s", 
				configpath ? configpath : "", /* avoid NULL */
				short_path(gf->title, gf->shift));
		sprintf(dlpath, "%s/pending/%s", configpath ? configpath : "", /* avoid NULL */
			short_path(gf->title, gf->shift));
		FUNCPRINT("moveing %s to %s", dlpath, finalpath);
		/* get rid of the nasty MS slashes */
		for(i = 0; dlpath[i] != '\0'; i++)
			if(dlpath[i] == '\\') dlpath[i]='/';
		/* make the directories for the final path */
		for (i = 0; finalpath[i] != 0; i++) {
			if (finalpath[i] == '\\' || finalpath[i] == '/') {
				/* handle subdirs */
				finalpath[i] = 0;
				mkdir(finalpath, 0777);
				/*FUNCPRINT("making dir '%s'\n", finalpath);*/
				finalpath[i] = '/';
			}
		}
		if (configpath) g_free(configpath);
		FUNCPRINT("Moving the file to its new home.");
		if(rename(dlpath, finalpath)<0) FUNCPRINT("error moving");
	} else {
		/* There was an error on the Download */
		gf->errormsg = strerror(errno);
		gf->status = EG_ERROR;
	}
	/* close the socket */
	if (sockfd != -1) close(sockfd);
	T( list_update( gtk_clist_find_row_from_data(GTK_CLIST(clist), gf)); )
	FUNCPRINT("The download completed successfully");
	downloadcomplete(gf);

/* close up the pthread exit handler */
pthread_cleanup_pop(0);

	return NULL;
}

static void queuedownload(shostqueue *hq) {
	GList *lqueue = hq->lqueue;
	sgetfile *gf;
	FUNCPRINT("host=%s", hq->host);
	while (lqueue != NULL) {
		gf = lqueue->data;
		if (gf->status == EG_READY) {
			/* file ready for download: pop it on the download queue */
			FUNCPRINT("queueing another download");
			downloadcount++;
			gf->status = EG_DOWNLOADING;
			hq->lqueue = g_list_remove_link(hq->lqueue, lqueue);
			hq->ldl = g_list_append(hq->ldl, gf);
			hq->downloadcur++;
			pthread_create(&gf->thread, NULL, downloadproc, gf);
			return;
		}
		lqueue = lqueue->next;
	}
}

static int queuefileready(shostqueue *hq) {
	GList *lqueue;
	sgetfile *gf;

	FUNCPRINT("host=%s", hq->host);
	lqueue = hq->lqueue; 
	if (lqueue == NULL) return FALSE;
	while (lqueue != NULL) {
		gf = lqueue->data;
		if (gf->status == EG_READY) 
		{
			FUNCPRINT("File Ready");
			return TRUE; /* one file is ready; return */
		}
		lqueue = lqueue->next;
	}
	/* if it gets here, then there are no "ready files";
	 * get info on the next available file, then return FALSE */
	lqueue = hq->lqueue; 
	while (lqueue != NULL) {
		if (gf->status == EG_QUEUED || gf->status == EG_REMOTEQUEUEFULL) {
			FUNCPRINT("trying to get info again");
			getinfo(gf);
			return FALSE;
		}
		lqueue = lqueue->next;
	}
	FUNCPRINT("return");
	return FALSE;
}

static void queuecheck(shostqueue *hq) {
	FUNCPRINT("host=%s", hq->host);
	TLOCK(hq->mutex);
	
	while (queuefileready(hq) &&          /* stuff on the queue and */
			(g_list_length(hq->ldl) < hq->downloadmax) /* and download queue isn't full */
			) {
		if (downloadcount > gnome_config_get_int(CONFIG_PREFIX "/Network/MaxGlobalDL=1")) 
		{
			TUNLOCK(hq->mutex);
			return;
		}
		/* add a download */
		FUNCPRINT("lqueue: %d, ldl: %d", g_list_length(hq->lqueue), g_list_length(hq->ldl));
		queuedownload(hq);
	} 
	TUNLOCK(hq->mutex);
}

static void queuecheckall() {
	GList *hl;
	shostqueue *hq;

	FUNCPRINT("(no args)");

	G_LOCK(hostlist);
	hl = hostlist;
	
	while (hl != NULL) {
		hq = hl->data;
		FUNCPRINT("Next Host");
		queuecheck(hq);
		hl = hl->next;
	}
	G_UNLOCK(hostlist);
}
/***************************************
static shostqueue*
hostqueue_find(char *host, int newok)

Find a particular host name in the list of
hosts 

host: the name of the host to find
newok: TRUE make a new entry if it isn't found
	FALSE don't

RETURNS: pointer to the host entry that was
	either created or found.  NULL if it wasn't
	created or found.
***************************************/
static shostqueue* 
hostqueue_find(char *host, int newok) {
	GList *l;
	shostqueue *lhq = NULL;
	g_assert(host!=NULL);
	FUNCPRINT("host='%s' newok=%d", host, newok);
	/* search host list for host */
	G_LOCK(hostlist);
	/*l = hostlist;*/
	for(l = hostlist; l != NULL; l = l->next) {
		lhq = l->data;
		g_assert(lhq != NULL);
		TLOCK(lhq->mutex);
		if (strcmp(lhq->host, host) == 0) {
			FUNCPRINT("found host '%s' on list.", lhq->host);
			TUNLOCK(lhq->mutex);
			break;
		}
		FUNCPRINT("it isn't host %s...", lhq->host);
		TUNLOCK(lhq->mutex);
		/*l = l->next;*/
	}
	if (l == NULL && newok) { /* no download proc for this host */
		FUNCPRINT("host '%s' not found, adding...", host);
		lhq = shostqueue_new(host);
		hostlist = g_list_append(hostlist, lhq);
	}
	G_UNLOCK(hostlist);
	return lhq;
}
/****************************************************
static sgetfile*
getfile_find(char *host, char *title, int newok)

Search the queue list for the specified host and title

host: the host name to search 
title: the filename to search for
newok:  if TRUE the function will make a new entry if 
	a match isn't found

RETURNS: pointer to the structure found
*****************************************************/
static sgetfile*
getfile_find(char *host, char *title, int newok) {
	GList *l;
	sgetfile *gf = NULL;
	shostqueue *lhq = hostqueue_find(host, newok);

	FUNCPRINT("host='%s' title='%s' newok=%d", host, title, newok);
	if (lhq == NULL) return NULL;

	TLOCK(lhq->mutex);
	l = lhq->lqueue;
	while (l != NULL) {
		gf = l->data;
		if (strcmp(gf->title, title) == 0) {
			FUNCPRINT("found it!");
			break;
		}
		l = l->next;
	}
	if (l == NULL && newok) {
		gf = sgetfile_new(title);
		gf->hq = lhq;
		lhq->lqueue = g_list_append(lhq->lqueue, gf);
	}
	else if(l == NULL && !newok)
	{
		FUNCPRINT("Didn't find it");
		gf = NULL;
	}
	TUNLOCK(lhq->mutex);
	
	return gf;
}
/***************************************************
static sgetfile*
getfile_find_download(char *host, char *title) 

search the download queue for the specified file

host: the host for the file
title: the name of the file

RETURNS: pointer to the struct that was found to match
***************************************************/
static sgetfile* getfile_find_download(char *host, char *title) {
	GList *l;
	sgetfile *gf = NULL;
	shostqueue *lhq = hostqueue_find(host, FALSE);

	FUNCPRINT("host='%s' title='%s'", host, title);
	if (lhq == NULL) return NULL;

	TLOCK(lhq->mutex);
	l = lhq->ldl;
	while (l != NULL) {
		gf = l->data;
		if (strcmp(gf->title, title) == 0) {
			FUNCPRINT("found it!");
			break;
		}
		l = l->next;
	}
	TUNLOCK(lhq->mutex);
	return gf;
}
/***********************************************************
static void download_check_hosts_valid(void)

check the lists of hosts.  If a host entry has no nodes in its queues
it is removed.
***********************************************************/
static void download_check_hosts_valid(void)
{
	GList *l = hostlist;
	GList *temp = NULL;
	shostqueue *hq;
	G_LOCK(hostlist);
	while(l != NULL)
	{
		hq = l->data;
		g_assert(hq != NULL);
		if(!g_list_length(hq->ldl) && !g_list_length(hq->lqueue))
		{
			if(hq->mutex)
				g_mutex_free(hq->mutex);
			hostlist = g_list_remove_link(hostlist, l);
			temp = l;
			l = l->next;
			g_list_free(temp);
		}
		else l = l->next;
			
	}
	G_UNLOCK(hostlist);
}
/***********************************************************
static void download_cancel_cb(GtkWidget *w, gpointer d)

Callback for canceling downloads
***********************************************************/
static void download_cancel_cb(GtkWidget *w, gpointer d)
{
	int row;
	GList *sel;
	int selected;
	sgetfile *gf;
	gboolean worked = FALSE;
	GList *newhead = NULL;
	shostqueue *hq;
	FUNCPRINT("Canceling Download");
	/* we'll be using the clist */
	G_LOCK(clist);
 	sel = GTK_CLIST(clist)->selection;
	if(sel == NULL) 
	{
		G_UNLOCK(clist);
		return;	
	}
	gf = gtk_clist_get_row_data(GTK_CLIST(clist), (int)sel->data); 
	g_assert(gf != NULL);
	hq = gf->hq;


	/* send the cancel request to the thread */
	if(gf->thread)
	{
		FUNCPRINT("Destroying thread");
		/* there is a thread created */
		pthread_cancel(gf->thread);
	}
	else
	{
		FUNCPRINT("Removing queue entry");
		newhead = download_remove_from_queue(gf, hq->lqueue, &worked);

		if(worked) hq->lqueue = newhead;
		else g_assert_not_reached();
	}
	/* remove the entry from the clist */
	gtk_clist_remove(GTK_CLIST(clist), selected=(int)sel->data); 
	/* and keep the row selected */
	gtk_clist_select_row(GTK_CLIST(clist), selected,0);
	/* now we are done with the clist */
	G_UNLOCK(clist);

	download_check_hosts_valid();
/* now the thread just has to check for a cancel request! */
}
/*************************************************************
static void download_remove_complete_cb(GtkWidget *w, gpointer d)

Callback to remove all entries in the clist that are marked as "File Complete".
*************************************************************/
static void download_remove_complete_cb(GtkWidget *w, gpointer d)
{
	int row;
	sgetfile *gf;
	shostqueue *hq;
	gboolean worked=FALSE;
	GList *newhead = NULL;
	G_LOCK(clist);
	for(row=0;; row++)
	{
		if((gf = gtk_clist_get_row_data(GTK_CLIST(clist), row))==NULL)
			break;
		hq = gf->hq;
		if(gf->status == EG_FILECOMPLETE) 
		{
			FUNCPRINT("Removing queue entry");
			newhead = download_remove_from_queue(gf, hq->ldl, &worked);

			if(worked) hq->lqueue = newhead;
			else g_assert_not_reached();
			

			/* remove the entry from the clist */
			gtk_clist_remove(GTK_CLIST(clist), row); 
		}
	} 
	G_UNLOCK(clist);
}
/***********************************************************
static void button_press_cb(GtkWidget *w, GdkEventButton *e, gpointer d)

Show the popup menu
***********************************************************/
static gint button_press_cb(GtkWidget *w, GdkEventButton *e, gpointer d) {
	GtkWidget *menu;
	sgetfile *gf;
	int row, col;

	if (e->type == GDK_BUTTON_PRESS && e->button == 3) {
		gtk_clist_get_selection_info(GTK_CLIST(clist), e->x, e->y,
				&row, &col);
		gf = gtk_clist_get_row_data(GTK_CLIST(clist), row);
		if (gf == NULL) return FALSE;
		gtk_clist_select_row(GTK_CLIST(clist), row, col);
		menu = gnome_popup_menu_new(popup_menu);
		gnome_popup_menu_do_popup_modal(menu, NULL, NULL, e, gf->hq->host);
		gtk_widget_destroy(menu);
		return TRUE;
	}

	return FALSE;
}
/***********************************************************
static void host_info_cb(GtkWidget *w, gpointer d)

Display a dialog with information about the host
***********************************************************/
static void host_info_cb(GtkWidget *w, gpointer d) {
        view_host_info(d);
}
/***********************************************************
GtkWidget* downloadtab_create(void)

Make the download tab
***********************************************************/
GtkWidget* downloadtab_create(void) {
	GtkWidget *scroll;
	GtkWidget *vbox, *hbox;
	GtkWidget *b;

	vbox = gtk_vbox_new(FALSE, 10);
/* the buttons on top */
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
		hbox = gtk_hbox_new(FALSE, 10);
		/* Make the button for canceling downloads */
			b = gtk_button_new_with_label("Cancel");
		gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 0);
		gtk_signal_connect(GTK_OBJECT(b), "clicked", 
				GTK_SIGNAL_FUNC(download_cancel_cb), NULL);
		/* Make the button to cancel completed downloads */
			b = gtk_button_new_with_label("Remove Complete");
		gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 0);
		gtk_signal_connect(GTK_OBJECT(b), "clicked",
				GTK_SIGNAL_FUNC(download_remove_complete_cb), NULL);
		gtk_widget_set_sensitive(hbox, TRUE);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, FALSE, FALSE, 0);

/* the clist on bottom */
		scroll = gtk_scrolled_window_new(NULL, NULL);
		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), 
				GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
			clist = gtk_clist_new_with_titles(sizeof(titles)/sizeof(char*), titles);
			/*gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_MULTIPLE);
			gtk_clist_set_column_justification(GTK_CLIST(clist), COL_SIZE,
					GTK_JUSTIFY_RIGHT);
			gtk_clist_set_column_justification(GTK_CLIST(clist), COL_BITRATE,
					GTK_JUSTIFY_RIGHT);
			*/
			gtk_clist_set_column_width(GTK_CLIST(clist), COL_TITLE, 300);
			gtk_clist_set_column_width(GTK_CLIST(clist), COL_HOST, 50);
			/*gtk_clist_set_column_width(GTK_CLIST(clist), COL_SIZE, 60);
			gtk_clist_set_column_width(GTK_CLIST(clist), COL_HOST, 60);
			gtk_clist_set_column_width(GTK_CLIST(clist), COL_BITRATE, 40);*/

			gtk_signal_connect(GTK_OBJECT(clist), "button_press_event",
				GTK_SIGNAL_FUNC(button_press_cb), NULL);

		gtk_container_add(GTK_CONTAINER(scroll), clist);
	gtk_box_pack_start(GTK_BOX(vbox), scroll, TRUE, TRUE, 0);

	//gtk_timeout_add(DOWNLOAD_UPDATE_FREQ, kbsec_calc_all, NULL);
	return vbox;
}
