/* gnome-napster, a GNOME Napster client.
 * Copyright (C) 1999 Evan Martin <eeyem@u.washington.edu>
 */
#include <gtk/gtk.h>
#include <gnome.h>
#include <time.h>
#include "songinfo.h"
#include "gtkflist.h"
#include "host.h"
#include "ui.h"
#include "lnap.h"
#include "gconfig.h"
#include "util.h"
#include "download.h"

#define CLIENT_LENGTH 40
#define HOST_CONFIG "/Host"
#define MAX_KEYS   999
#define KEY_LENGTH 64
typedef struct 
{
	char name[HOST_LENGTH];
	gboolean status;
} userinfo;

static void host_info_cb(GtkWidget *w, gpointer d);
static void host_remove_hostinfo_cb(GtkWidget *w, gpointer d);

static GnomeUIInfo popup_menu[] = {
        GNOMEUIINFO_ITEM_NONE(N_("_Remove from host list"),
                        N_("Remove this host from the Host tab"),
                        host_remove_hostinfo_cb),
        GNOMEUIINFO_ITEM_NONE(N_("_Host Info"),
                        N_("View Information about this host"),
                        host_info_cb),
        GNOMEUIINFO_END
};
/*****************************************
Globals
*****************************************/
static GtkWidget *clist, *hostlist, *controlbox;
static GList *songlist = NULL;
static int songcount = 0;
#define  STATUS_ACTIVE_STR "Active"

enum {
	COL_NAME, COL_SIZE, COL_BITRATE, COL_TIME, COL_COUNT
};
static gchar* clist_titles[] = { "Song Name", "Size", "kbps", "Length" };

/*****************************************
Prototypes
*****************************************/
int host_find_user_entry(char *name);
void host_remove_hostinfo(char *name);


/*****************************************
Functions
*****************************************/
static void browsehost(char *host) {
	char *add[COL_COUNT];

	songlist = g_list_free_all(songlist);
	songcount = 0;

	memset(add, 0, sizeof(add));
	add[0] = g_strdup_printf("Requesting file list from '%s'...", host);
	gtk_clist_clear(GTK_CLIST(clist));
	gtk_clist_append(GTK_CLIST(clist), add);
	g_free(add[0]);

	nap_sendpacket(nap_socknap, NM_BROWSE_REQUEST, strlen(host), host);
}

static void browse_cb(GtkWidget *w, gpointer d) {
	char *host;
	GList *sel;

	sel = GTK_CLIST(hostlist)->selection;
	if (sel == NULL) return;
	gtk_clist_get_text(GTK_CLIST(hostlist), (int)sel->data, 0, &host);

	browsehost(host);
}

static void
list_update(int row) {
	ssong_info *i = gtk_clist_get_row_data(GTK_CLIST(clist), row);
	char *text = short_path(i->title, i->pathshift);
	gtk_clist_set_text(GTK_CLIST(clist), row, COL_NAME, text);
}

static void 
list_add(ssong_info *i) {
	gchar* add[COL_COUNT];
	int row;

	add[COL_NAME] = short_path(i->title, i->pathshift);
	add[COL_SIZE] = g_strdup_printf("%lu", i->size);
	add[COL_TIME] = g_strdup_printf("%u:%02u", i->time/60, i->time % 60);
	add[COL_BITRATE] = g_strdup_printf("%d", i->bitrate);
	row = gtk_clist_append(GTK_CLIST(clist), add);
	gtk_clist_set_row_data(GTK_CLIST(clist), row, i);
};

static void list_statustext(char *text) {
	char *add[COL_COUNT];
	memset(add, 0, sizeof(add));
	add[0] = text;
	gtk_clist_clear(GTK_CLIST(clist));
	gtk_clist_append(GTK_CLIST(clist), add);
}

void host_notfound(char *text) {
	char *add = g_strdup_printf("Host '%s' not found.", text);
	list_statustext(add);
	g_free(add);
}

void host_result(char *text) {
	ssong_info *i;
	char buf[50];

	i = songinfo_new();
	sscanf(text, "%s \"%[^\"]\" %s %lu %d %d %d",
			i->host, i->title, i->checksum, &i->size, 
			&i->bitrate, &i->samprate, &i->time);

	i->pathshift = gnome_config_get_int(CONFIG_PREFIX "/Interface/PathDepth=1");

	songcount++;
	snprintf(buf, 50, "Got %d results...", songcount);
	gtk_clist_set_text(GTK_CLIST(clist), 0, COL_NAME, buf);
	
	songlist = g_list_append(songlist, i);
}

void host_complete(void) {
	GList *l = songlist;

	list_statustext("Updating list...");
	gtk_clist_freeze(GTK_CLIST(clist));
	gtk_clist_clear(GTK_CLIST(clist));
	while (l != NULL) {
		list_add(l->data);
		l = l->next;
	}
	gtk_clist_thaw(GTK_CLIST(clist));
}
/**********************************
static void host_info_cb(GtkWidget *w, gpointer d)

called from the popup menu, shows info about the host
**********************************/
static void host_info_cb(GtkWidget *w, gpointer d)
{
/*
	nap_sendpacket(nap_socknap, NM_WHOIS_REQUEST, strlen((char *)d), d);
*/
	view_host_info(d);
}
static void host_info_dialog_destroy(GtkWidget *w, gpointer d)
{
	gtk_widget_destroy(GTK_WIDGET(d));
}
/************************************
static void host_info_dialog(int message, char *info)

Creates the dialog to show info
	message: the message that the server returned
		WHOIS or WHOWAS
	info: the info returned by the server
	
	Nothing returned
************************************/
void host_info_dialog(int message, char *info)
{
	GtkWidget *dialog;
	GtkWidget *vbox;
	GtkWidget *label;
	char *info_format[8]={"Name: %s",
				"User Level: %s",
				"Time Connected: %s",
				"Status: %s",
				"Files Shared: %s",
				"Current Downloads: %s",
				"Current Uploads: %s",
				"Client: %s"};
	char data[8][20];
		/*<nick>
		<user-level> is one of "User" or "Admin"
	       	<time> is seconds this user has been connected
	        <status> is one of "Active." or "Inactive." if they are on or offline
	        <shared> is number of files user has available for download
	        <downloads> is the current number of downloads in progress
	        <uploads> is the current number of uploads in progress
	        <client-info> see 0x02 (client login) above
		*/
	char channels[256];
	char temp_str[264];
	int conn_type;
	int index;
	time_t last_seen;
/*	dialog = gnome_dialog_new("Host Information", GNOME_STOCK_BUTTON_OK,
NULL);*/
	dialog = gnome_ok_dialog("Host Information");
	vbox=gtk_vbox_new(TRUE, 5);
	/* figure out the format of the message from the NM_* message type */
	if(message==NM_WHOIS_RESPONSE)
	{
		/* the user is still online */
		/* take apart the message into its peices */
		sscanf(info, "%s \"%[^\"]\" %s \"%[^\"]\" \"%[^\"]\" %s %s %s %d \"%[^\"]\"",
			data[0], data[1], data[2], 
			channels, data[3], data[4], data[5], data[6], 
			&conn_type, data[7]);
		/* create all of the labels */
		for(index=0; index<8; index++)
		{
			/* handle splitting up the time into a more 
				readable form */
			if(index==2)
			{
				sscanf(data[index], "%ld", &last_seen);
				g_snprintf(temp_str, sizeof(temp_str),
					"%ld:%ld:%ld", last_seen/3600, (last_seen%3600)/60, last_seen%60);
				strcpy(data[index], temp_str);
			}
			g_snprintf(temp_str, sizeof(temp_str), info_format[index], data[index]);
			label=gtk_label_new(temp_str);
			/* set the justification to the left */
			gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
			gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 0);
		}	
		g_snprintf(temp_str, sizeof(temp_str), "Connection: %s", nap_conntypes[conn_type]);
		label=gtk_label_new(temp_str);
		/* Justify to the left */
		gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
		gtk_box_pack_start(GTK_BOX(vbox), label, TRUE, TRUE, 0);

		/* and put it into the dialog */
		gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dialog)->vbox), vbox, FALSE, FALSE, 0);
	}
	else if(message==NM_WHOWAS_RESPONSE)
	{
		/* they are now offline */

		sscanf(info, "%s \"%[^\"]\" %ld", data[0], data[1], &last_seen);
		g_snprintf(temp_str, sizeof(temp_str), info_format[0], data[0]);
		label=gtk_label_new(temp_str);
		/* justify to the left */
		gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
		gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);

		g_snprintf(temp_str, sizeof(temp_str), info_format[1], data[1]);
		label=gtk_label_new(temp_str);
		/* justify to the left */
		gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
		gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
		
		g_snprintf(temp_str, sizeof(temp_str), "Last seen:\t%s", ctime(&last_seen));
		label=gtk_label_new(temp_str);
		/* justify to the left */
		gtk_misc_set_alignment(GTK_MISC(label), 0, 0.5);
		gtk_box_pack_start(GTK_BOX(vbox), label, FALSE, FALSE, 0);
		
		/* and put it into the dialog */
		gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(dialog)->vbox), vbox, FALSE, FALSE, 0);
	}
	gnome_dialog_button_connect(GNOME_DIALOG(dialog), 0,
		GTK_SIGNAL_FUNC(host_info_dialog_destroy), dialog);
	gtk_widget_show_all(dialog);
}
/**************************************
static void host_remove_hostinfo_cb(GtkWidget *w, gpointer d)

Called from the popup menu to remove a host from the list
***************************************/
static void host_remove_hostinfo_cb(GtkWidget *w, gpointer d)
{
	host_remove_hostinfo((char *)d);
}
/***********************************
void host_remove_hostinfo(char *name)

Remove a user(host) entry from the hostinfo
list.
	name: the name of the host to remove
***********************************/
void host_remove_hostinfo(char *name)
{
	int user;
	userinfo *info;
	if((user=host_find_user_entry(name))==-1)
	{
		/* it wasn't in the list */
		return;
	}
	info=gtk_clist_get_row_data(GTK_CLIST(hostlist), user);
	gtk_clist_set_row_data(GTK_CLIST(hostlist), user, NULL);
	nap_sendpacket(nap_socknap, NM_HOTLIST_REMOVE, strlen(info->name), info->name);
	g_free(info);
	gtk_clist_remove(GTK_CLIST(hostlist), user);
}
/******************************************
void host_addhost(char *name)

Add a host to the host window for easy browsing

	name: the name of the host(user) to add
******************************************/
void host_addhost(char *name) {
	char *add[1] = { name };
	int row;
	userinfo *info;

	if(host_find_user_entry(name)!=-1) return;
	row = gtk_clist_append(GTK_CLIST(hostlist), add);

	info = g_malloc(sizeof(userinfo));
	strcpy(info->name, name);
	gtk_clist_set_row_data(GTK_CLIST(hostlist), row, info);
	host_user_status(FALSE, name);
	/* now need to tell the server that we have a new person on the hotlist */
	nap_sendpacket(nap_socknap, NM_HOTLIST_ADD, strlen(name), name);
}
/**********************************
void view_host_info(gpointer d)

called to show info about a host
**********************************/
void view_host_info(gpointer d)
{
	nap_sendpacket(nap_socknap, NM_WHOIS_REQUEST, strlen((char *)d), d);
}
/**************************************
void host_init_hostlist(void)

Called at startup _after_ the connection is made, load the hotlist in from
the gnome-config database
**************************************/
void host_init_hostlist(void)
{
	gchar *value;
	char key[KEY_LENGTH];
	int index=0;
	do
	{	
		/* construt the key name */
		g_snprintf(key, KEY_LENGTH, "%s%s/%d", CONFIG_PREFIX, HOST_CONFIG, index+1); 

		value = gnome_config_get_string(key); /* get the value of the key */
		if(value==NULL) break;
		host_addhost(value); /* add it to the list */

		g_free(value); 
	} while(index++ <= MAX_KEYS);
}
/**********************************
void host_save_hostlist(void)

Called at shutdown to save the hotlist
**********************************/
void host_save_hostlist(void)
{
	userinfo *data;
	char key[KEY_LENGTH];
	int index=0;
	
	g_snprintf(key, KEY_LENGTH,"%s%s", CONFIG_PREFIX, HOST_CONFIG);
	
	gnome_config_clean_section(key);
	gnome_config_sync();

	while(index <= MAX_KEYS)
	{
		data=gtk_clist_get_row_data(GTK_CLIST(hostlist), index);
		if(data != NULL)
		{	
			
			/* construct the key */
			g_snprintf(key, KEY_LENGTH,"%s%s/%d", CONFIG_PREFIX, HOST_CONFIG, index+1);
			/* save just the name since we can figure everything else
				out from it */
			gnome_config_set_string(key, data->name);
			/* move on to the next element */
		}
		else break;
	index++;
	}
	gnome_config_sync();
}
/****************************************
GList *host_find_user_entry(char *name)

Search through the hostinfo list looking
for name.
	name: the name to look for
	RETURNS: the row the name is on
****************************************/
int host_find_user_entry(char *name)
{	
	int index;
	userinfo *info;
	for(index=0;;index++)
	{

		info=gtk_clist_get_row_data(GTK_CLIST(hostlist), index);
		if(info==NULL) break;
		
		if(!strcmp(info->name, name))
		{
			/* the host has been found in the list */
			return index;
		}
	}
	return -1;
}
/**************************************
void host_user_added(gboolean worked, char *user_name)

Called when the server sends a response to a request to
add a user to the hotlist.
	worked: TRUE if the user was added
		FALSE if the user wasn't
	user_name: the name of the user affected
**************************************/
void host_user_added(gboolean worked, char *user_name)
{
	/*TODO: add a message to let the user know 
		the status of the add */
	int row;
	userinfo *user;
	if(worked)
	{
		if((row=host_find_user_entry(user_name))!=-1)
		{
			user=gtk_clist_get_row_data(GTK_CLIST(hostlist), row);	
			/* put SOMETHING here! */
		}
	}
	else
	{
		/* it didn't work */
		/* so take it out of the hostinfo list */
		FUNCPRINT("couldn't add host %s", user_name);
		host_remove_hostinfo(user_name);
	}
}
/*********************************************
void host_user_status(gboolean status, char *user_name)

	user_name: the name of the user to change
	status: the status to change user_name to
		TRUE==online
		FALSE==offline
**********************************************/
void host_user_status(gboolean status, char *user_name)
{
	int row;
	char user[40];
	int speed;
	GdkColor red={128, 65535, 0, 0};
	GdkColor green={128, 0, 65535, 0};	
	userinfo *user_data;
	sscanf(user_name, "%s %d", user, &speed);
	if((row=host_find_user_entry(user))==-1)
	{
		/* he's not in the list */
		/* let's add 'im */
		host_addhost(user);
		return;
	
	}
	user_data=gtk_clist_get_row_data(GTK_CLIST(hostlist), row);
	user_data->status=status;
	gtk_clist_set_foreground(GTK_CLIST(hostlist), row, status ? &green : &red);
}
/* enable or disable the Browse Host button */
void host_login_status(int loggedin) {
	static gboolean was_loggedin=FALSE;
	gtk_widget_set_sensitive(controlbox, loggedin);
	if(loggedin) {host_init_hostlist(); was_loggedin=TRUE;}
	else if(!loggedin && was_loggedin) { host_save_hostlist();
was_loggedin=FALSE; }
}

/* implement column sorting when selecting column headings */
static void 
sort_col(GtkCList *clist, gint column, gint width, gpointer user_data)
{
	switch (column) {
		case COL_SIZE:
			gtk_clist_set_compare_func(clist, songinfo_compare_size);
			break;
		case COL_BITRATE:
			gtk_clist_set_compare_func(clist, songinfo_compare_bitrate);
			break;
		case COL_TIME:
			gtk_clist_set_compare_func(clist, songinfo_compare_time);
			break;
		case COL_NAME:
		default:
			gtk_clist_set_compare_func(clist, NULL); /* default compare func */
	}

	if (column == clist->sort_column) {
		/* switch sort type */
		gtk_clist_set_sort_type(clist, GTK_SORT_DESCENDING + GTK_SORT_ASCENDING
				- clist->sort_type);
	} else {
		/* select column to sort */
		gtk_clist_set_sort_column(clist, column);
	}

	/* go for it */
	gtk_clist_sort(clist);
}
/*******************************
static void getfiles_cb()

Called when the user double clickes on the file list widget
Adds the clicked file to the download queue.
called from button_press_cb()
	No args taken
	Nothing returned
*******************************/
static void getfiles_cb() {
	GList *sel;
	ssong_info *i;

	sel = GTK_CLIST(clist)->selection;
	if (sel == NULL) return;
	while (sel != NULL) {
		i = gtk_clist_get_row_data(GTK_CLIST(clist), (int)sel->data);
		if (i != NULL) /* NULL possible if they select "no matches found" */
			download_enqueue(i->host, i->title, i->pathshift);
		sel = sel->next;
	}
}

static void shift_foreach_cb(gpointer data, gpointer d) {
	int row = (int)data;
	int shift = (int)d;
	ssong_info *i = gtk_clist_get_row_data(GTK_CLIST(clist), row);

	if (i == NULL) return;

	if (i->pathshift + shift < 0) i->pathshift = 0;
	else i->pathshift += shift;
	list_update(row);
}

static void shift_cb(GtkWidget *w, gpointer d) {
	if (GTK_CLIST(clist)->selection)
		g_list_foreach(GTK_CLIST(clist)->selection, shift_foreach_cb, d);
}
/*************************************************************************
static gint filelist_button_press_cb(GtkWidget *w, GdkEventButton *e, gpointer d)

handle button presses in the file list widget
	GtkWidget: the widget that this is being called for
	GdkEventButton: button press info
	gpointer: just some extra info, isn't used here

RETURNS:
	TRUE: if it handled the button press
	FALSE: if it didn't handle the button press
*************************************************************************/
static gint filelist_button_press_cb(GtkWidget *w, GdkEventButton *e, gpointer d) {

	if (e->type == GDK_2BUTTON_PRESS) {
		getfiles_cb();
		return TRUE;
	}
	return FALSE;
}
/*************************************************************************
static gint hostlist_button_press_cb(GtkWidget *w, GdkEventButton *e, gpointer d)

handle button presses in the host list widget
	GtkWidget: the widget that this is being called for
	GdkEventButton: button press info
	gpointer: just some extra info, isn't used here

RETURNS:
	TRUE: if it handled the button press
	FALSE: if it didn't handle the button press
*************************************************************************/
static gint hostlist_button_press_cb(GtkWidget *w, GdkEventButton *e, gpointer d) {
	GtkWidget *menu;
	userinfo *info;
	int row, col;
	if (e->type == GDK_BUTTON_PRESS && e->button == 3) {
		if(gtk_clist_get_selection_info(GTK_CLIST(hostlist), e->x, e->y, &row, &col))
		{
			info=gtk_clist_get_row_data(GTK_CLIST(hostlist), row);
			if(info==NULL) return FALSE;
			menu = gnome_popup_menu_new(popup_menu);
			gnome_popup_menu_do_popup_modal(menu, NULL, NULL, e, info->name);
			gtk_widget_destroy(menu);
			return TRUE;
		}
	}
	else if (e->type == GDK_2BUTTON_PRESS)
	{
		browse_cb(w, d);
		return TRUE;
	}
	return FALSE;
}



GtkWidget*
create_host(void) {
	GtkWidget *mainbox, *paned;
	GtkWidget *hostvbox, *scroll;
	GtkWidget *filevbox;
	GtkWidget *hbox, *b;

	static gchar *host_title[] = { "Host" };
	mainbox = gtk_vbox_new(FALSE, 10);
	gtk_container_set_border_width(GTK_CONTAINER(mainbox), 10);

	paned = gtk_hpaned_new();
	gtk_paned_set_gutter_size(GTK_PANED(paned), 20);
/* make the left box containing the list of hosts and browse button */
	hostvbox = gtk_vbox_new(FALSE, 10);
		/* note: this is a flist, not a standard clist */
		hostlist = gtk_flist_new_with_titles(1, host_title);
	gtk_box_pack_start(GTK_BOX(hostvbox), hostlist, TRUE, TRUE, 0);
		controlbox = gtk_vbox_new(FALSE, 10);
			b = gtk_button_new_with_label("Browse Host");
			gtk_signal_connect(GTK_OBJECT(b), "clicked",
					GTK_SIGNAL_FUNC(browse_cb), NULL);
		gtk_box_pack_start(GTK_BOX(controlbox), b, FALSE, FALSE, 0);
		/* handle the right mouse button to pop up the menu */
		 gtk_signal_connect(GTK_OBJECT(hostlist), "button_press_event",
                        GTK_SIGNAL_FUNC(hostlist_button_press_cb), NULL);
	gtk_box_pack_start(GTK_BOX(hostvbox), controlbox, FALSE, FALSE, 0);

/* make the file list box on the right side */
	filevbox = gtk_vbox_new(FALSE, 10);
	scroll = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), 
			GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
		/* note that this is also an flist, not a clist */
		clist = gtk_flist_new_with_titles(sizeof(clist_titles)/sizeof(char*), 
				clist_titles);
		hbox = gtk_hbox_new(FALSE, 0);
				b = songinfo_arrow_button_new(app, TRUE);
				gtk_signal_connect(GTK_OBJECT(b), "clicked",
						GTK_SIGNAL_FUNC(shift_cb), (gpointer)-1);
			gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 0);
				b = songinfo_arrow_button_new(app, FALSE);
				gtk_signal_connect(GTK_OBJECT(b), "clicked",
						GTK_SIGNAL_FUNC(shift_cb), (gpointer)1);
			gtk_box_pack_start(GTK_BOX(hbox), b, FALSE, FALSE, 0);

			gtk_box_pack_start(GTK_BOX(hbox), gtk_label_new(clist_titles[0]), 
					FALSE, FALSE, 5);
		gtk_widget_show_all(hbox);
		gtk_clist_set_column_widget(GTK_CLIST(clist), 0, hbox);
		gtk_clist_set_selection_mode(GTK_CLIST(clist), GTK_SELECTION_MULTIPLE);
		gtk_clist_set_column_justification(GTK_CLIST(clist), COL_SIZE,
				GTK_JUSTIFY_RIGHT);
		gtk_clist_set_column_justification(GTK_CLIST(clist), COL_BITRATE,
				GTK_JUSTIFY_RIGHT);
		gtk_clist_set_column_justification(GTK_CLIST(clist), COL_TIME,
				GTK_JUSTIFY_RIGHT);
		gtk_clist_set_column_width(GTK_CLIST(clist), COL_NAME, 230);
		gtk_clist_set_column_width(GTK_CLIST(clist), COL_SIZE, 50);
		gtk_clist_set_column_width(GTK_CLIST(clist), COL_BITRATE, 25);
		gtk_signal_connect(GTK_OBJECT(clist), "button_press_event",
				GTK_SIGNAL_FUNC(filelist_button_press_cb), NULL);
		gtk_signal_connect(GTK_OBJECT(clist), "click-column",
				GTK_SIGNAL_FUNC(sort_col), NULL);
		gtk_clist_set_sort_column(GTK_CLIST(clist), 0);	/* sort on title by default */
		gtk_clist_set_auto_sort(GTK_CLIST(clist), TRUE);
	gtk_container_add(GTK_CONTAINER(scroll), clist);
	gtk_box_pack_start(GTK_BOX(filevbox), scroll, TRUE, TRUE, 0);
		hbox = gtk_hbox_new(FALSE, 10);
			/* Make the Get File(s) button */
			b = gtk_button_new_with_label("Get File(s)");
			gtk_signal_connect(GTK_OBJECT(b), "clicked",
				GTK_SIGNAL_FUNC(getfiles_cb), NULL);
		gtk_box_pack_end(GTK_BOX(hbox), b, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(filevbox), hbox, FALSE, FALSE, 0);
/* put the two parts in the host panel */
	gtk_paned_add1(GTK_PANED(paned), hostvbox);
	gtk_paned_add2(GTK_PANED(paned), filevbox);

	gtk_box_pack_start(GTK_BOX(mainbox), paned, TRUE, TRUE, 0);
	return mainbox;	
}


