/* gnome-napster, a GNOME Napster client.
 * Copyright (C) 1999 Evan Martin <eeyem@u.washington.edu>
 */
#include <gtk/gtk.h>
#include <errno.h>
#include <gnome.h>
#include <pthread.h>
#include <fcntl.h>
#include <glib.h>
#include <zvt/zvtterm.h>
#include "ui.h"
#include "util.h"
#include "lnap.h"
#include "login.h"
#include "search.h"
#include "host.h"
#include "download.h"
#include "gconfig.h"

/* from configure.in */
#include "config.h"

#include "../pixmaps/napsmall.xpm"
#include "../pixmaps/online.xpm"
#include "../pixmaps/offline.xpm"

void about_cb(GtkWidget *w, gpointer d);
static void init_login(GtkWidget *w, gpointer d);
void log_complete(int type);

GtkWidget *app;
static GtkWidget *term, *appbar, *connpix, *statuslabel;
static gint gtag_alt_download = 0;
static gint gtag_server = 0;
int alt_sock = 0;

/* colors "borrowed" from crescendo/src/color.c :) */
static gushort blu[] =
{
	0x0000,0x0000,0x0000,0x0000,0xaaaa,0xaaaa,0xaaaa,0xaaaa,
	0x5555,0x5555,0x5555,0x5555,0xffff,0xffff,0xffff,0xffff,
	0xaaaa, 0x0000
	/* 0xd7d7, 0x0000 */
};
static gushort red[] =
{
	0x0000,0xaaaa,0x0000,0xaaaa,0x0000,0xaaaa,0x0000,0xaaaa,
	0x5555,0xffff,0x5555,0xffff,0x5555,0xffff,0x5555,0xffff,
	0xaaaa, 0x0000
	/* 0xfafa, 0x0000 */
};
static gushort grn[] =
{ 
	0x0000,0x0000,0xaaaa,0x5555,0x0000,0x0000,0xaaaa,0xaaaa,
	0x5555,0x5555,0xffff,0xffff,0x5555,0x5555,0xffff,0xffff,
	0xaaaa, 0x0000
	/* 0xebeb, 0x0000 */
}; 

GnomeUIInfo file_menu[] = {
	GNOMEUIINFO_MENU_EXIT_ITEM(gtk_main_quit, NULL),
	GNOMEUIINFO_END
};

GnomeUIInfo napster_menu[] = {
	GNOMEUIINFO_ITEM(N_("_Login/Logout"), N_("Login/logout to Napster"), 
			init_login, napsmall_xpm),
	GNOMEUIINFO_ITEM_NONE(N_("_Relog"), N_("Logout, then log back in to Napster"), 
			login_relog_cb),
	GNOMEUIINFO_END
};

GnomeUIInfo settings_menu[] = {
	GNOMEUIINFO_MENU_PREFERENCES_ITEM(gconfig_dlg_run, NULL),
	GNOMEUIINFO_END
};

GnomeUIInfo help_menu[] = {
	GNOMEUIINFO_MENU_ABOUT_ITEM(about_cb, NULL),
	GNOMEUIINFO_END
};

GnomeUIInfo menubar[] = {
	GNOMEUIINFO_MENU_FILE_TREE(file_menu),
	GNOMEUIINFO_SUBTREE(N_("_Napster"), napster_menu),
	GNOMEUIINFO_MENU_SETTINGS_TREE(settings_menu),
	GNOMEUIINFO_MENU_HELP_TREE(help_menu),
	GNOMEUIINFO_END
};

static void delete_event(GtkWidget *w, GdkEvent *e, gpointer d) {
	gtk_main_quit();
}

inline void printc(char *text) {
	zvt_term_feed(ZVT_TERM(term), text, strlen(text));
}

inline void printcn(char *text) {
	printc(text); printc("\r\n");
	qprintf("%s\n", text);
}

void printcf(char *fmt, ...) {
	char *text;
	va_list ap;
	
	va_start(ap, fmt);
	text = g_strdup_vprintf(fmt, ap);
	va_end(ap);
	
	printc(text); qprintf(text);
	g_free(text);
}


static void status_cb(char *text, void *d) {
	printcn(text);
}

static void enter_cb(GtkWidget *w, gpointer d) {
	GtkEntry *e = GTK_ENTRY(d);
	gchar *text;
	
	text = gtk_entry_get_text(e);
	printcn(text); 
	gtk_entry_set_text(e, "");
}

static GtkWidget* create_console() {
	GtkWidget *vbox, *hbox, *e, *sb;
	vbox = gtk_vbox_new(FALSE, 10);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
		hbox = gtk_hbox_new(FALSE, 0);
			term = zvt_term_new_with_size(80, 24);
			zvt_term_set_blink(ZVT_TERM(term), FALSE);
			zvt_term_set_scrollback(ZVT_TERM(term), 1000);
			GTK_WIDGET_UNSET_FLAGS(term, GTK_CAN_FOCUS);
		gtk_box_pack_start(GTK_BOX(hbox), term, TRUE, TRUE, 0);
			sb = gtk_vscrollbar_new(GTK_ADJUSTMENT(ZVT_TERM(term)->adjustment));
			GTK_WIDGET_UNSET_FLAGS(sb, GTK_CAN_FOCUS);
		gtk_box_pack_start(GTK_BOX(hbox), sb, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(vbox), hbox, TRUE, TRUE, 0);
		e = gtk_entry_new();
		gtk_signal_connect(GTK_OBJECT(e), "activate",
				GTK_SIGNAL_FUNC(enter_cb), e);
	gtk_box_pack_start(GTK_BOX(vbox), e, FALSE, FALSE, 0);
	nap_setstatusfunc(status_cb, NULL);
	return vbox;
}

static void init_login(GtkWidget *w, gpointer d) {
	/*gtk_widget_set_sensitive(menubar[1].widget, FALSE);*/
	login_loginout_cb(NULL, NULL); /* popup login box */
}
/****************************
void
ui_run(void)

Called from main().  Sets up the application 
then enters the gtk_main() loop.
****************************/
void 
ui_run(void) {
	GtkWidget *book;
	gint screen_x;
	gint screen_y;

	app = gnome_app_new(PACKAGE, _("GNOME Napster Client " VERSION));
	gnome_app_create_menus_with_data(GNOME_APP(app), menubar, app);
	gtk_window_set_policy(GTK_WINDOW(app), TRUE, TRUE, FALSE);
	/* get the screen size */
	screen_x = gdk_screen_width();
	screen_y = gdk_screen_height();
	/* by default make us half the screen dimensions */
	gtk_window_set_default_size(GTK_WINDOW(app), screen_x/2, screen_y/2);
	gtk_signal_connect(GTK_OBJECT(app), "delete_event", 
		GTK_SIGNAL_FUNC(delete_event), NULL);

	appbar = gnome_appbar_new(FALSE, TRUE, GNOME_PREFERENCES_USER);
	gnome_appbar_set_default(GNOME_APPBAR(appbar), _("Ready"));
	gnome_app_set_statusbar(GNOME_APP(app), appbar);

	statuslabel = gtk_label_new("");
	gtk_box_pack_start(GTK_BOX(appbar), statuslabel, FALSE, FALSE, 5);

	connpix = gnome_pixmap_new_from_xpm_d(offline_xpm);

	gtk_box_pack_start(GTK_BOX(appbar), connpix, FALSE, FALSE, 0);

	gnome_app_install_menu_hints(GNOME_APP(app), menubar);
	
	gtk_widget_realize(app);

	book = gtk_notebook_new();
	gtk_notebook_set_scrollable(GTK_NOTEBOOK(book), TRUE);
	gtk_notebook_append_page(GTK_NOTEBOOK(book), 
			create_console(), gtk_label_new("Console"));
	gtk_notebook_append_page(GTK_NOTEBOOK(book), 
			create_search(), gtk_label_new("Search"));
	search_login_status(FALSE);
	gtk_notebook_append_page(GTK_NOTEBOOK(book), 
			create_host(), gtk_label_new("Host"));
	host_login_status(FALSE);
	/* gtk_notebook_append_page(GTK_NOTEBOOK(book), create_queue(), gtk_label_new("Queue")); */
	gtk_notebook_append_page(GTK_NOTEBOOK(book), 
			downloadtab_create(), gtk_label_new("Download"));

	gnome_app_set_contents(GNOME_APP(app), book);

	gtk_widget_show_all(app);
	zvt_term_set_color_scheme(ZVT_TERM(term), red, grn, blu);
	
	gdk_threads_enter();
	/* gtk_idle_add(init_login, NULL);
	 * can't add it as an idle process, because it pops up a modal dialog. */

	init_login(NULL, NULL);
	gtk_main();
	gdk_threads_leave();
	host_login_status(FALSE);
}
/*****************************************************************************
static inline GtkWidget *about_url_new(char *link, char *text, char *caption)

Helper function for about_cb.  Makes the url links in the about box.
*****************************************************************************/
static inline GtkWidget *about_url_new(char *link, char *text, char *caption) {
	GtkWidget *hb;
	GtkWidget *url, *l;
	hb = gtk_hbox_new(FALSE, 0);
		url = gnome_href_new(link, text); 
	gtk_box_pack_start(GTK_BOX(hb), url, FALSE, FALSE, 0);
		l = gtk_label_new(caption);
		gtk_label_set_justify(GTK_LABEL(l), GTK_JUSTIFY_RIGHT);
		gtk_label_set_line_wrap(GTK_LABEL(l), TRUE);
	gtk_box_pack_end(GTK_BOX(hb), l, FALSE, FALSE, 0);
	return hb;
}
/*******************************************
void about_cb(GtkWidget *w, gpointer d)

Make the about dialog
*******************************************/
void about_cb(GtkWidget *w, gpointer d) {
	static GtkWidget *about = NULL;
	const gchar *authors[] = {
		"Andrew Parker",
		"Evan Martin",
		NULL
	};
	GtkWidget *vbox;
	GtkWidget *url;
	gchar *logofile;
	
	if ((logofile = gnome_pixmap_file("gnome-napster-largelogo.png")) == NULL) {
		g_warning("gnome-napster-animate.png not in gnome datadir; trying"
				" a relative path.");
		logofile = "../pixmaps/gnome-napster-largelogo.png";
	} else {
		g_free(logofile);
		logofile = "gnome-napster-largelogo.png";
	}
	about = gnome_about_new(_("GNOME Napster Client"), VERSION, 
			"Copyright (C) 2000 Evan Martin and Andrew Parker",
			authors, _("Client for Napster, a music sharing system."), 
			logofile);

	gnome_dialog_set_parent(GNOME_DIALOG(about), GTK_WINDOW(app));

	vbox = gtk_vbox_new(FALSE, 5);
	gtk_container_set_border_width(GTK_CONTAINER(vbox), 10);
	gtk_box_pack_start(GTK_BOX(vbox), gtk_label_new("Relevant web links:"), FALSE, FALSE, 0);
	url = about_url_new("http://gnome-napster.sourceforge.net", 
				"Gnome Napster", "Home of this program.");
	gtk_box_pack_start(GTK_BOX(vbox), url, TRUE, TRUE, 0);
	url = about_url_new("http://www.napster.com", 
			"www.napster.com", "Home of Napster.");
	gtk_box_pack_start(GTK_BOX(vbox), url, TRUE, TRUE, 0);
	url = about_url_new("http://www.napster.com/faq", 
			"Napster FAQ", "Many of these FAQs also\napply to this client.");
	gtk_box_pack_start(GTK_BOX(vbox), url, TRUE, TRUE, 0);

	gtk_box_pack_start(GTK_BOX(GNOME_DIALOG(about)->vbox), vbox, FALSE, FALSE, 0);
	gtk_widget_show_all(GNOME_DIALOG(about)->vbox);

	/*gtk_window_set_modal(GTK_WINDOW(about), TRUE);*/
	gnome_dialog_run_and_close(GNOME_DIALOG(about));
}
/********************************************
static void handlemsg(snap_packet *p)

Handle the messages from the server, routing them to the
appropriate function

p:  the struct containing the packet info
********************************************/
static void handlemsg(snap_packet *p) {
	int a, b, c;
	char buf[128];

	switch (p->msg) {

	/* searching control */
		case NM_ERROR:
			gnome_error_dialog_parented(p->txt, GTK_WINDOW(app));
			break;
		case NM_SEARCH_RESULT: 
			search_result(p->txt);
			break;
		case NM_SEARCH_COMPLETE:
			search_complete();
			break;

	/* File browseing control (in the host window) */
	/* This is the same numeric value for two different messages */
		/*case NM_HOTLIST_USER_OFF:*/
		case NM_BROWSE_NOTFOUND: 
			host_notfound(p->txt);
			host_user_status(FALSE, p->txt); 
			break;
		case NM_BROWSE_RESULT:
			host_result(p->txt);
			break;
		case NM_BROWSE_COMPLETE:
			host_complete();
			break;
	/* File download */
		case NM_FILEINFO_RESULT:
			download_getinfo_cb(p->txt);
			break;
		case NM_REMOTEQUEUEFULL:
			nap_dumppacket(p);
			download_remotequeuefull_cb(p->txt);
			break;
		case NM_FILECOUNT:
			sscanf(p->txt, "%d %d %d", &a, &b, &c);
			sprintf(buf, "%d songs in %d libraries (%dgb) on %s", 
					b, a, c, g_logininfo.serv);
			printcn(buf);
			gtk_label_set_text(GTK_LABEL(statuslabel), buf);
			break;

	/* Hotlist control */	
		case NM_HOTLIST_ACK:
			host_user_added(TRUE, p->txt);
			break;
		case NM_HOTLIST_ERROR:
			host_user_added(FALSE, p->txt);
			break;
		case NM_HOTLIST_USER_ON:
			host_user_status(TRUE, p->txt);
			break;
		case NM_SYSMSG: /* system message */
			printcn(p->txt);
			break;
		case NM_WHOIS_RESPONSE:
			host_info_dialog(NM_WHOIS_RESPONSE, p->txt);
			break; 
		case NM_WHOWAS_RESPONSE:
			host_info_dialog(NM_WHOWAS_RESPONSE, p->txt);
			break;

		default:
			nap_dumppacket(p);
			break;
	}
	free(p);
}
/****************************************************************
static void alternate_download_input(gpointer d, gint source, GdkInputCondition condition)

The input callback for the alternate download port.
****************************************************************/
static void alternate_download_input(gpointer d, gint source, GdkInputCondition condition)
{
	char recv_buf[512];
	snap_packet *p;
	char host[40];
	char filename[256];
	unsigned long filelength;
	char ack = '1';
	int sock, len;
	struct sockaddr_in s;
	int size = sizeof(s);
	int flags;
	FUNCPRINT("We got called");

	/* accept the connection */
	if((sock = accept(alt_sock, &s, &size)) < 0)
	{
		FUNCPRINT("Error on socket accept");
		return;
	}
	FUNCPRINT("Got Connection!");	
	/* send the ack */
	if(send(sock, &ack, sizeof(ack), 0) < 0) {
		perror("send");
		return;
	}
	FUNCPRINT("Recving response");

	/* RECV request */
	if((len = recv(sock, recv_buf, 512, 0)) < 0)
	{
		FUNCPRINT("didn't work");
		close(sock);
		return;
	}
	recv_buf[len]= '\0';
	if(!strcmp(recv_buf, "SEND"))
	{
		/* FIXME: this will hang the program for a while
			if the remote client doensn't respond immediatly. 
			Maybe put the new socket on its own gdk_input handler
			until we have all the info we need? */
		/* RECV what they are sending */
		if((len = recv(sock, recv_buf, 512, 0)) < 0)
		{
			FUNCPRINT("didn't work");
			close(sock);
			return;
		}

		recv_buf[len]='\0';
		FUNCPRINT("sent %s", recv_buf);
		/* get the info */
		sscanf(recv_buf, "%s \"%[^\"]\" %u", host, filename, &filelength);
		/* release the thread */
		FUNCPRINT("Host = '%s' Filename = '%s' Length = %u", host, filename, filelength);
		/* change the socket to Non blocking */
		/*flags = fcntl(sock, F_GETFL);

		if(fcntl(sock, F_SETFL, (flags ^ O_NONBLOCK)) < 0)
		{
			FUNCPRINT("Error setting socket options");
			close(sock);
			return;
		}*/

		if(!download_release(sock, host, filename, filelength))
		{
			/* oops couldn't find it */
			close(sock);
			return;
		}
	}
	else if(!strcmp(recv_buf, "GET"))
	{
		/* handle the upload request */
		close(sock); /* remove this line when this is done */
		return;
	}
	else
	{
		FUNCPRINT("Unknown response from remote client: %s", recv_buf);
		close(sock);
	}
}
/*************************************************************
static void
server_input(gpointer d, gint source, GdkInputCondition condition)

The input callback for the connection to the server
*************************************************************/
#define MAX_FAIL 20
static void
server_input(gpointer d, gint source, GdkInputCondition condition) {
	snap_packet *p;
	/*FIXME: this failed connection check is kinda a hack */
	static int fail_count = 0;
	p = nap_readpacket_nonblock(nap_socknap);
	if (p == NULL) 
	{
		/*FUNCPRINT("Error reading socket: %s(%d)",strerror(errno), errno);*/
		fail_count++;
		/* check to see if it is really an error */
		if(fail_count > MAX_FAIL)
		{
			/* we've failed to many times */
			/* log us out */
			nap_logout(); 
			g_logininfo.connected = FALSE;
			ui_loginstatus();
		}
		return;	
	}
	else fail_count = 0;
	handlemsg(p);
}

/*static void* dataproc(void *d) {
	char buf[1024];
	int len, i, sock;
	int sock;
	struct sockaddr_in addr;
	int addr_size = sizeof(struct sockaddr_in);
	if (nap_socklocal == -1) return NULL;
	while (1) {
		if ((sock = accept(nap_socklocal, (struct sockaddr*)&addr, &addr_size)) < 0) {
			perror("accept");
			continue;
		}
		qprintf("accepted connection from %s...\n", inet_ntoa(addr.sin_addr));
	}
	return NULL;
}*/
/***********************************
extern void ui_loginstatus(void)

Handle all of the changes for when we change status(on/off line)
***********************************/
extern void ui_loginstatus(void) {
	if (g_logininfo.connected) {
		/* login complete */
		/* load the logged in picture */
		gnome_pixmap_load_xpm_d(GNOME_PIXMAP(connpix), online_xpm);
		/* add the input handler for the server */
		gtag_server = gdk_input_add(nap_socknap, GDK_INPUT_READ, server_input, NULL);
		/* and the stuff for alternate downloads */
		if((alt_sock=download_init()) > 0)
		{
			FUNCPRINT("made socket, now connecting handler");
			gtag_alt_download = gdk_input_add(alt_sock, 
				GDK_INPUT_READ, 
				alternate_download_input, NULL);
		}
		/* tell the various parts that we are on */
		search_login_status(TRUE);
		host_login_status(TRUE);
	} else {
		/* load the not logged in picture */
		gnome_pixmap_load_xpm_d(GNOME_PIXMAP(connpix), offline_xpm);
		qprintf("killing server/data threads...\n");
		/* remove the input handler */
		gdk_input_remove(gtag_server);
		gdk_input_remove(gtag_alt_download);
		download_shutdown();
		gtag_alt_download = 0;
		gtag_server = 0;
		printcn("Logged out.");
		/* tell the various parts we are off */
		search_login_status(FALSE);
		host_login_status(FALSE);
	}
}
